## RANDOM
## This file implements Box domain WITH GMP

## Copyright 2008-2012 Gianluca Amato, Francesca Scozzari

## This file is part of RANDOM: R-based Analyzer for Numerical DOMains
##
## RANDOM is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## RANDOM is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty ofa
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with RANDOM.  If not, see <http://www.gnu.org/licenses/>.

box.domain.gmp <- function() {	
	
	pmin <- pmin.bigqi
	pmax <- pmax.bigqi
	
	## Generate new boxes from scratch
	
	full <- function(n=1) { 
		list(size=n, min=as.bigqi(c(rep(-Inf,n),1)), max=as.bigqi(c(rep(+Inf,n),1))) 
	}
	
	empty <- function(n=1) {
		list(size=n, min=as.bigqi(c(rep(+Inf,n),1)), max=as.bigqi(c(rep(-Inf,n),1)))
	}
	
	zero <- function(n=1) {
		list(size=n, min=as.bigqi(c(rep(0,n),1)), max=as.bigqi(c(rep(0,n),1)))
	}
	
	frombox <- function(box) {    
		box$min <- as.bigqi(box$min)
		box$max <- as.bigqi(box$max)
		box
	}
	
	getsize <- function(box) {
		box$size    
	}
	
	## Relations between boxes
	
	is.subseteq <- function(box1,box2) {
		all(box1$min >= box2$min) && all(box1$max <= box2$max)
	}
	is.equal <- function(box1,box2) {
		all(box1$min == box2$min) && all(box1$max == box2$max)
	}
	is.subset <- function(box1,box2) {
		is.subseteq(box1,box2) && (! is.equal(box1,box2))
	}
	is.empty <- function(box) {
		any(box$min > box$max)
	}
	
	## Min / ArgMin
	
	argmin <- function (box, linear.form)  {
		ifelse.bigqi(linear.form >= 0, box$min, box$max)  
	}
	
	argmax <- function (box, linear.form)  {
		ifelse.bigqi(linear.form >=0, box$max, box$min)  
	}
	
	minimize <- function (box, linear.form) {
		min.bounds <- argmin(box,linear.form)
		result <- dot.product(as.bigqi(linear.form), as.bigqi(min.bounds))
		return(result)
	}
	
	maximize <- function (box, linear.form) {
		max.bounds <- argmax(box,linear.form)
		result <- dot.product(as.bigqi(linear.form), as.bigqi(max.bounds))
		return(result)
	}  
		
	## Operations
	
	assign <- function(box, index, linear.form, newparam=NULL) {
		if (is.empty(box))
			box
		else {
			newbox <- box
			newbox$min[index] <- minimize(box,linear.form)
			newbox$max[index] <- maximize(box,linear.form)
			newbox
		}
	}
	
	assign.unknown <- function(box, index) {
		if (is.empty(box))
			empty(box$size)
		else {
			box$min[index] <- as.bigqi(-Inf)
			box$max[index] <- as.bigqi(+Inf)
			box
		}
	}
	
	select <- function(box, linear.form) {
		lf.argmin <- argmin(box,linear.form)
		linear.form <- as.bigqi(linear.form)
		lf.min <- dot.product(linear.form, lf.argmin)
		if (lf.min > 0)
			empty(box$size)
		else {
			outbox.min <- box$min
			outbox.max <- box$max
			
			pos.inf <- which(is.infinite(lf.argmin) & linear.form!=0)
			if (length(pos.inf)==0) {
				val <- pmax(box$min, lf.argmin -lf.min/linear.form)     
				outbox.min[linear.form<0] <- val[linear.form<0]
				val <- pmin(box$max, lf.argmin -lf.min/linear.form)     
				outbox.max[linear.form>0] <- val[linear.form>0]
			} else if (length(pos.inf)==1) {
				if (linear.form[pos.inf]<0) {
					outbox.min[pos.inf] <- max(box$min[pos.inf],
							- dot.product(as.bigqi(linear.form[-pos.inf]),
									lf.argmin[-pos.inf])/as.vector(linear.form[pos.inf]))
				} else {
					outbox.max[pos.inf] <- min(box$max[pos.inf],
							- dot.product(as.bigqi(linear.form[-pos.inf]),
									lf.argmin[-pos.inf])/as.vector(linear.form[pos.inf]))
				}
			}
			list(size=box$size,min=outbox.min,max=outbox.max)      
		}
	}
	
	select.diff <- function(box, linear.form) {
		lf.argmin <- argmin(box,linear.form)
		lf.min <- dot.product(linear.form, lf.argmin)
		
		if (lf.min != 0) 
			box
		else if (all((linear.form==0) | (box$max==box$min)))
			empty(box$size)
		else
			box
	}
	
	union <- function(box1,box2) {
		list(size=box1$size, min = pmin(box1$min,box2$min), max=pmax(box1$max,box2$max))
	}
	
	intersection <- function(box1,box2) {
		min <- pmax(box1$min,box2$min)
		max <- pmin(box1$max,box2$max)
		
		if (any(min > max))
			empty(box1$size)
		else
			list(size=box1$size, min=min, max=max)
	}
	
	## Widening/Narrowing
	
	widening <- function(box1,box2,delay=0) {
		if (is.empty(box1))
			box2
		else {
			if (delay>0) 
				result <- union(box1,box2)
			else {
				box1$min <- ifelse.bigqi(box2$min < box1$min,as.bigqi(-Inf), box1$min)
				box1$max <- ifelse.bigqi(box2$max > box1$max,as.bigqi(+Inf), box1$max)
				result <- box1
			}			
			result
		}
	}
	
	narrowing <- function(box1,box2) {
		if (is.empty(box1))
			box1
		else {
			box1$min <- ifelse.bigqi(box1$min == as.bigqi(-Inf), box2$min, pmin(box1$min, box2$min))
			box1$max <- ifelse.bigqi(box1$max == as.bigqi(+Inf), box2$max, pmax(box1$max, box2$max))
			box1
		}
	}
	
	## Output
	
	as.character <- function(box,vars=paste("v",1:box$size,sep="")) {
		if (is.empty(box)) return ("empty")
		out <- ""
		for (i in 1:length(vars)) {
			if ((box$min[i]==-Inf) && (box$max[i]==+Inf)) 
				newout <- ""
			else if (box$min[i]==-Inf)
				newout <- paste(vars[i],"<=",box$max[i],sep="")
			else if (box$max[i]==+Inf)
				newout <- paste(box$min[i],"<=",vars[i], sep="")
			else if (box$max[i]==box$min[i])
				newout <- paste(vars[i],"=",box$max[i], sep="")
			else
				newout <- paste(box$min[i],"<=",vars[i],"<=",box$max[i], sep="")
			out <- paste(out,newout,sep="")
			if (i < length(vars) && newout!="")
				out <- paste(out," , ",sep="")
		}
		out
	}
	
	setparam <- function(box, v) {
		box
	}	
	
	outputform <- function(box,lf,vars=paste("v",1:box$size,sep="")) {
		min.bound <- minimize(box,lf)
		max.bound <- maximize(box,lf)		
		as.character(list(size=1, min=c(min.bound,1), max=c(max.bound,1)), linear.form.display(lf,vars))
	}
	
	## Declaratation of public functions
	
	list (full = full, empty = empty, zero=zero, frombox=frombox, getsize=getsize,
			is.subseteq = is.subseteq, is.subset = is.subset, is.equal = is.equal, is.empty = is.empty,
			argmin = argmin, argmax = argmax, minimize = minimize, maximize = maximize,
			assign = assign, assign.unknown = assign.unknown, 
			select = select, select.diff = select.diff,
			union = union, intersection = intersection,
			widening = widening, narrowing=narrowing, 
			as.character=as.character, setparam=setparam, outputform=outputform)
}
