## This file implements the domain of parallelotopes WITH GMP

## Copyright 2008-2012 Gianluca Amato, Francesca Scozzari

## This file is part of RANDOM: R-based Analyzer for Numerical DOMains
##
## RANDOM is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## RANDOM is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty ofa
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copyof the GNU General Public License
## along with RANDOM.  If not, see <http://www.gnu.org/licenses/>.

generalptopes.domain.gmp <- function(aggregates) {		
	
	GCD.THRESHOLD <- as.bigq(1,1000)

	DEBUG <-  FALSE

	# Simplifies a constraints by removing the gcd	
	simplify.row <- function(v) {
		v/gcd(v, zero.threshold = GCD.THRESHOLD)
	}
	
	boxdom <- box.domain.gmp()
	
	# rot is the matrix of rays
	# irot is the matrix of constraints 
	
	check <- function(rbox) {
		if (! identical(length(rbox$max),length(rbox$min))) 
			stop("Bounds of parallelotopes should have the same size")
		if ( rbox$size != length(rbox$min) )
			stop("Bounds of parallelotopes should have the same size as the environment dimension")
		if (any(is.na(rbox$min)) || any(is.na(rbox$max))) 
			stop("Bounds should not contain NA or NaN")
		if (any(is.na(rbox$rot)) || any(is.na(rbox$irot)))
			stop("Rotation matrix should not contain NA or NaN")
		if (! identical(as.vector(rbox$irot[rbox$size,]), as.bigq(c(rep(0,rbox$size-1),1) ))) 
			stop("Normalization is bugged")	
						
		center <- as.bigqi(head(rbox$irot[,rbox$size],-1))
		rbox$min <- rbox$min - c(center,0)
		rbox$max <- rbox$max - c(center,0)
		rbox$irot[,rbox$size] <- c(rep(0,rbox$size - 1),1)
		rbox$rot[,rbox$size] <- c(rep(0,rbox$size - 1),1)
		rbox
	}
	
	new <- function(min, max, rot, irot, expand=FALSE) {
		size <- length(min)
		if (missing(rot) && missing(irot)) {		
			rot <- as.bigq(diag(size))			
			irot <- rot			
		} else if (missing(irot)) {
			rot <- as.bigq(rot)
			irot <- solve.trick(irot)		
		} else if (missing(rot)) {
			irot <- as.bigq(irot)				
			rot <- solve.trick(irot)
		}
		min <- as.bigqi(min)
		max <- as.bigqi(max)
		if (expand) {
			min <- c(min,1)
			max <- c(max,1)			
			# rbind is bugged in gmp 0.5.. it neeeds a matrix as second argument
			rot <-  rbind(cbind(rot, rep(0, size)), t(c(rep(0,size),1)))			
			irot <- rbind(cbind(irot, rep(0, size)), t(c(rep(0,size),1)))
			size <- size + 1
		}
		check( list(size=size, rot=rot, irot=irot, min=min, max=max ) )
	}	
	
	new.simple <- function(min, max, irot) {
		size <- length(min)
		if (missing(irot))			
			irot <- diag(size)
		else
			irot <- matrix(irot, ncol=size, byrow=TRUE)
		new(min, max, irot=irot, expand=TRUE)
	}
	
	full <- function(n=1, rot, irot, expand=TRUE) {
		new( rep(-Inf,n), rep(+Inf,n), rot, irot, expand)		
	}
	
	empty <- function(n=1, rot, irot, expand=TRUE) {	
		new( rep(+Inf,n), rep(-Inf,n), rot, irot, expand)		
	}
	
	zero <- function(n=1, rot, irot, expand=TRUE) {			
		new( rep(0,n), rep(0,n), rot, irot, expand )
	}
	
	frombox <- function(box) {
		new( box$min, box$max, expand=FALSE )		
	}
	
	tobox <- function(rbox) {
		list(size=rbox$size-1, min=rbox$min, max=rbox$max)	
	}
	
	getsize <- function(rbox) {
		rbox$size - 1
	}
	
	newrot <- function(rbox, newirot) {
		newrbox <- full( rbox$size, irot=newirot, expand=FALSE )
		irot <- as.bigq(newirot) %*% rbox$rot
		for (i in 1:newrbox$size) {
			newrbox$min[i] <- boxdom$minimize(tobox(rbox),irot[i,])
			newrbox$max[i] <- boxdom$maximize(tobox(rbox),irot[i,])
		}		
		check( newrbox )
	}
	
	## Relations between boxes (apporximate versions)
	
	is.subseteq <- function(rbox1,rbox2) {
		rbox1rot <- newrot(rbox1,rbox2$irot)
		all(rbox1rot$min >= rbox2$min) && all(rbox1rot$max <= rbox2$max)
		
	}
	
	is.equal <- function(rbox1,rbox2) {		
		is.subseteq(rbox1,rbox2) && is.subseteq(rbox2,rbox1)
	}
	
	is.subset <- function(rbox1,rbox2) {
		is.subseteq(rbox1,rbox2) && (! is.subseteq(rbox2,rbox1))
	}
	
	is.empty <- function(rbox) {
		any(rbox$min > rbox$max)
	}	
	
	argmin <- function () {}
	argmax <- function () {}
	
#   # Minimization / Maximization
#	
#   # do not work with infinite bounds, but apply does not work with a result of bigqi
#	
#	argmin <- function (rbox, linear.form)  {
#		linear.form <- as.bigq(t(linear.form)) %*% rbox$rot		
#		min.bounds <- apply(rbox$rot,2,function(x) as.bigq(sum(ifelse.bigqi(linear.form >= 0, rbox$min, rbox$max) * as.bigqi(x))))
#		head(min.bounds,-1)
#	}
#	
#	argmax <- function (rbox, linear.form)  {
#		linear.form <- as.bigq(t(linear.form)) %*% rbox$rot	
#		max.bounds <- apply(rbox$rot,2,function(x) as.bigq(sum(ifelse.bigqi(linear.form >= 0, rbox$max, rbox$min) * as.bigqi(x))))
#		head(max.bounds,-1)		 
#	}
#	
	minimize <- function (rbox, linear.form) {
		linear.form <- as.bigq(t(linear.form)) %*% rbox$rot  
		dot.product.gmp(as.bigqi(linear.form), ifelse.bigqi(linear.form >= 0, rbox$min, rbox$max))		
	}
	
	maximize <- function (rbox, linear.form) {
		linear.form <- as.bigq(t(linear.form)) %*% rbox$rot  
		dot.product.gmp(as.bigqi(linear.form), ifelse.bigqi(linear.form >= 0, rbox$max, rbox$min))		
	}  
	
	## Operations
	
	assign <- function(rbox, index, linear.form, newparam=NULL) {
		if (is.empty(rbox)) return(rbox)
		if (linear.form[index] == 0) {
			rbox <- assign.unknown(rbox,index)
			candidates <- which(rbox$irot[,index] != 0)			
			ordering <- c(setdiff(1:nrow(rbox$irot),candidates),candidates)
			newrow <- simplify.row(replace(-linear.form, index, 1))
			topivot <- t(rbind( t(newrow) ,rbox$irot[ordering,])) 			
			# we convert topivot to double since qr does not work in gmp. The use of t(newrow) is needed
			# for bugs in rbind.gmp
			topivot.double <- matrix(as.numeric(topivot), nrow=nrow(topivot), ncol=ncol(topivot))
			pivot <- qr(topivot.double)$pivot			
			chosen <- ordering[[tail(pivot,1)-1]]
			rbox$irot[chosen, ] <- newrow 
			rbox$rot <- solve.trick(rbox$irot)
			rbox$min[chosen] <- 0
			rbox$max[chosen] <- 0
			check(rbox)
		} else {		
			oldx <- replace(-linear.form, index, 1)/linear.form[index]
			#rbox$irot <- rbox$irot - rbox$irot[,index] %*% t(replace(rep(0,rbox$size),index,1)) + rbox$irot[,index] %*% t(oldx)
		    rbox$irot <- rbox$irot - rbox$irot[,index] %*% t(replace(rep(0,rbox$size),index,1)) + rbox$irot[,index] %*% t(oldx)			
		
			# Simplify matrix by removing GCD
			rowsToBeModified <- which(rbox$irot[,index] != 0 )			
		    for (i in rowsToBeModified) {
				gcd <- gcd(as.vector(rbox$irot[i, ]))
				if (gcd > GCD.THRESHOLD) {
					rbox$irot[i,] <- rbox$irot[i,]/gcd
					rbox$min[i] <- rbox$min[i]/as.bigqi(gcd)
					rbox$max[i] <- rbox$max[i]/as.bigqi(gcd)
				}
			}
			
			rbox$rot <- solve.trick(rbox$irot)
			check(rbox) 
		}
	}
	
	assign.unknown <- function(rbox, index) {
		if (is.empty(rbox)) return(rbox)
		# Look for a good candidate for extracting the variable I want to remove
		candidates <- which(rbox$irot[,index]!=0 & rbox$min!=-Inf & rbox$max!=+Inf)
		if (length(candidates)==0) {
			candidates <- which(rbox$irot[,index]!=0 & (rbox$min!=-Inf | rbox$max!=+Inf))
		}
		if (length(candidates)==0) {
			# Very bad, not a good candidate
			# the following line are useless
			#rbox$min = ifelse(rbox$irot[,index]==0, rbox$min, -Inf)
			#rbox$max = ifelse(rbox$irot[,index]==0, rbox$max, +Inf)
			return (rbox)	
		}
		chosen <- candidates[1]
		multi.factor <-  rbox$irot[,index] / as.vector(rbox$irot[chosen,index])
		newirot <- rbox$irot - multi.factor %*% rbox$irot[chosen,]
		newmin <- rbox$min - as.bigqi(multi.factor) * ifsign.gmp(multi.factor, rbox$max[chosen], rbox$min[chosen])
		newmax <- rbox$max - as.bigqi(multi.factor) * ifsign.gmp(multi.factor, rbox$min[chosen], rbox$max[chosen])
		newirot[chosen,] <- rbox$irot[chosen,]
		newmin[chosen] <- -Inf
		newmax[chosen] <- +Inf
		new(newmin, newmax, irot=newirot)		
	}
	
	select <- function(rbox, linear.form) {				
		new.linear.form <- as.bigqi(as.bigq(t(linear.form)) %*% rbox$rot)
		newbox <- boxdom$select(tobox(rbox),new.linear.form)	
		chosen <- 0		
		candidates <- which(rbox$min == -Inf & rbox$max == + Inf & new.linear.form != 0)		
		if (length(candidates) != 0) 
			chosen <- candidates[1] 
		else {			
			candidates <- (rbox$min == -Inf | rbox$max == + Inf) & new.linear.form != 0						
			if (any(candidates)) {
				points <- ifelse.bigqi(candidates, ifelse.bigqi(rbox$min == -Inf, new.linear.form , -new.linear.form), -Inf)				
				# I would like to write chosen <- which.max(points), but which.max.gmp does not exist
				chosen <- head(which(max(points)==points),1)
			}				
		}		
		if (chosen != 0) {			
			newbox$min[chosen] <- dot.product.gmp(new.linear.form, ifelse.bigqi(new.linear.form >= 0, newbox$min, newbox$max))					
			newbox$max[chosen] <- min(as.bigqi(0),dot.product.gmp(new.linear.form, ifelse.bigqi(new.linear.form >= 0, newbox$max, newbox$min)))			
			rbox$irot[chosen,] <- linear.form	
		}				
		new(newbox$min, newbox$max, irot=rbox$irot)		
	}	
	
	select.diff <- function(rbox, linear.form) {
		new.linear.form <- as.bigqi(as.bigq(t(linear.form)) %*% rbox$rot)		
		newbox <- boxdom$select.diff(tobox(rbox),new.linear.form)	
		new(newbox$min, newbox$max, irot=rbox$irot)		
	}
	
	union.weak <- function(rbox1,rbox2) {
		if (is.empty(rbox1)) return(rbox2)
		if (is.empty(rbox2)) return(rbox1)		
		rbox2bis <- newrot(rbox2, rbox1$irot)
		new( pmin(rbox1$min, rbox2bis$min), pmax(rbox1$max, rbox2bis$max), irot=rbox1$irot)
	}
	
	union.inverse <- function(rbox1,rbox2) {
		
		if (DEBUG) {
			print(rbox1)
			print(rbox2)
		}
		
		linear.dep <- function(v1,v2) {
			v1 <- as.vector(v1)
			v2 <- as.vector(v2)
			chosen <- match(TRUE, v1!=0 & v2!=0)
			if (is.na(chosen)) 
				NA
			else 
			if (identical( v1/v1[chosen], v2/v2[chosen]))
				v1[chosen]/v2[chosen]
			else NA
		}			
		
		linear.dep.set <- function(v,constraints) {
			for (i in 1:nrow(constraints)) {
				if (! is.na(linear.dep(v[-size],constraints[i,-size]))) return(TRUE)
			}				
			return(FALSE)
		}
		
		new.constraint <- function(vi, vj, min1i, min2i, min1j, min2j) {
			if (is.infinite(min1i) | is.infinite(min2i) | is.infinite(min1j) | is.infinite(min2j))
				return(NULL)			
			vi <- as.bigqi(vi)
			vj <- as.bigqi(vj)
			deltai <- min1i-min2i
			deltaj <- min2j-min1j
			if (deltaj<0) { deltai <- -deltai; deltaj <- -deltaj }
			if (deltai * deltaj > 0) {
				simplify.row(as.bigq(c( -(deltaj*vi+deltai*vj), deltaj*min1i+deltai*min1j)))
			} else
				NULL
		}
		
		stable.constraints <- function(rboxa,rboxb) {
			# given two (part of) boxes on the same rotation matrix, computes the indexes
			# of the bounds which are stable or bistable
			bistable <- numeric(0)
			stable <- numeric(0)
			nostable <- numeric(0)
			for (i in 1:(size-1)) {							
				if (rboxa$min[i]==rboxb$min[i] & rboxa$max[i]==rboxb$max[i] & rboxa$min[i]!=-Inf & rboxa$max[i]!=+Inf)
					bistable <- c(bistable,i)		
				else if (rboxa$min[i]==rboxb$min[i] & rboxa$min[i] != -Inf)
					stable <- c(stable,i) 
				else if (rboxa$max[i]==rboxb$max[i] & rboxa$max[i] != +Inf)
					stable <- c(stable,i)
				else if (rboxa$max[i]!=+Inf | rboxa$min[i]!=-Inf)
					nostable <- c(nostable,i)
			}
			list(stable=stable, bistable=bistable, nostable=nostable)
		}
				
		if (is.empty(rbox1)) return(rbox2)
		if (is.empty(rbox2)) return(rbox1)
		size <- rbox1$size
		
		# Compute the rotated boxes
		
		rbox2b <- newrot(rbox2, rbox1$irot)
		rbox1b <- newrot(rbox1, rbox2$irot)
		
		# Identify stable constraints coming from rbox1
		sc1 <- stable.constraints(rbox1, rbox2b)
		
		# Identify rows in sc2 which are not linear dependent from rows in sc1
		
		goodrows <- 1:(size-1)
		for (i in 1:(size-1)) {
			found <- FALSE
			for (j in goodrows) 
				if (! is.na(linear.dep(rbox1$irot[i,],rbox2$irot[j,])) ) { found<- TRUE; break }
			if (found) goodrows <- setdiff(goodrows, j)	
		}					
		
		# Indentify bistable / stable / nostable constraints coming from rbox2
		
		sc2 <- stable.constraints(rbox2, rbox1b)
		sc2$stable <- intersect(sc2$stable, goodrows)
		sc2$bistable <- intersect(sc2$bistable, goodrows)
		sc2$nostable <- intersect(sc2$nostable, goodrows)
		
		# Initialize the bulk matrix.. I cannot create empty matrices or vectors because gmp is buggy
		
		bulk <- matrix(as.bigq(0),nrow=1, ncol=size)		
		min1 <- as.bigqi(0)
		min2 <- as.bigqi(0)
		max1 <- as.bigqi(0)
		max2 <- as.bigqi(0)
		
		# Put in bulk constraints coming from rbox1 which are not considered bistable
		for (i in c(sc1$nostable,sc1$stable)) {
			bulk <- rbind(bulk, rbox1$irot[i,])
			min1 <- c(min1,rbox1$min[i])
			min2 <- c(min2,rbox2b$min[i])
			max1 <- c(max1,rbox1$max[i])
			max2 <- c(max2,rbox2b$max[i])		
		}
		
		# Put in bulk constraints coming from rbox2 which are not bistable
		for (i in c(sc2$nostable,sc2$stable)) {
			bulk <- rbind(bulk, rbox2$irot[i,])
			min1 <- c(min1,rbox1b$min[i])
			min2 <- c(min2,rbox2$min[i])
			max1 <- c(max1,rbox1b$max[i])
			max2 <- c(max2,rbox2$max[i])					
		}		
		bulksize <- nrow(bulk)-1			
		
		# Normalize bulk and max/min
		min1 <- min1[-1]
		min2 <- min2[-1]
		max1 <- max1[-1]
		max2 <- max2[-1]
		bulk <- bulk[-1,]
				
		# Create new constraints 
		
		newconstraints <- matrix(as.bigq(0),nrow=1, ncol=size)
		newmax  <- numeric()		
				
		if (bulksize >= 2) {		
			for (i in 1:(bulksize-1))
				for (j in (i+1):bulksize) {
					nc <- new.constraint(bulk[i,-size],bulk[j,-size], min1[i], min2[i], min1[j], min2[j] )				
					if (! is.null(nc) && !linear.dep.set(nc,rbox1$irot) && !linear.dep.set(nc,rbox2$irot)) newconstraints <- rbind(newconstraints,t(nc))
					if (DEBUG && ! is.null(nc)) print(c(i,j,"mm"))
					nc <- new.constraint(bulk[i,-size],-bulk[j,-size], min1[i], min2[i], -max1[j], -max2[j] )
					if (! is.null(nc) && !linear.dep.set(nc,rbox1$irot) && !linear.dep.set(nc,rbox2$irot)) newconstraints <- rbind(newconstraints,t(nc))				
					if (DEBUG && ! is.null(nc)) print(c(i,j,"mM"))
					nc <- new.constraint(-bulk[i,-size],-bulk[j,-size], -max1[i], -max2[i], -max1[j], -max2[j] )					
					if (! is.null(nc) && !linear.dep.set(nc,rbox1$irot) && !linear.dep.set(nc,rbox2$irot)) newconstraints <- rbind(newconstraints,t(nc))
					if (DEBUG && ! is.null(nc)) print(c(i,j,"MM"))
					nc <- new.constraint(-bulk[i,-size], bulk[j,-size], -max1[i], -max2[i], min1[j], min2[j] )					
					if (! is.null(nc) && !linear.dep.set(nc,rbox1$irot) & !linear.dep.set(nc,rbox2$irot)) newconstraints <- rbind(newconstraints,t(nc))
					if (DEBUG && ! is.null(nc)) print(c(i,j,"Mm"))
				}	
		}
		dimnames(newconstraints)=NULL
		newconstraints <- newconstraints[-1,]
		# Remove not saturated constraints or constraints already existents
		
		saturated <- c()
		if (length(newconstraints)!=0)
			for (i in seq(1, length.out=nrow(newconstraints))) {
				min1 <- maximize(rbox1, t(newconstraints[i,]))
				min2 <- maximize(rbox2, t(newconstraints[i,]))
				if (min1==0 && min2==0) saturated <- c(saturated,i)
			}		
		
		# Put together compatible saturated constraints
		
		i <- 1
		bistable <- c()
		stable <- c()
		bistablemin <- as.bigqi(0)
		stablemin <- as.bigqi(0)		
		
		while (i <= length(saturated)) {
			tmpmin <- -Inf
			tmprm <- c()
			for (j in seq(i+1,length.out=length(saturated)-i)) {
				factor <- linear.dep(newconstraints[saturated[i],-size], newconstraints[saturated[j],-size])
				if (! is.na(factor)) {
					if (factor<0) tmpmin <- -newconstraints[saturated[j],size]/factor
					tmprm <- c(tmprm,j)
				}			
			}
			saturated <- setdiff(saturated, tmprm)
			if (as.bigqi(tmpmin) != -Inf) {
				bistable <- c(bistable,saturated[i])
				bistablemin <- c(bistablemin, as.bigqi(tmpmin))
			} else {				
				stable <- c(stable,saturated[i])
				stablemin <- c(stablemin,as.bigqi(min(minimize(rbox1,matrix(c(newconstraints[saturated[i],-size],0))), 
								minimize(rbox2,matrix(c(newconstraints[saturated[i],-size],0))))))
			}
			i <- i+1
		}		
		bistablemin <- bistablemin[-1]
		stablemin <- stablemin[-1]
		
		# Select best constraints in the order: bisaturated, saturated, others
		newirot <- matrix(as.bigq(0), ncol=size-1, nrow=1)
		newmin <- as.bigqi(0)
		newmax <- as.bigqi(0)
		
		if (length(sc1$bistable) !=0) {
			newirot <- rbind(newirot,rbox1$irot[sc1$bistable,-size])
			newmin <- c(newmin,rbox1$min[sc1$bistable])
			newmax <- c(newmax,rbox1$max[sc1$bistable])
		}
		if (length(sc2$bistable) !=0) {
			newirot <- rbind(newirot,rbox2$irot[sc2$bistable,-size])
			newmin <- c(newmin,rbox2$min[sc2$bistable])
			newmax <- c(newmax,rbox2$max[sc2$bistable])					
		}
		if (length(bistable) !=0) {
			newirot <- rbind(newirot, newconstraints[bistable,-size])  # I should remove constraints I have in rbox1 and rbox2		
			newmin <- c(newmin, bistablemin)
			newmax <- c(newmax, -newconstraints[bistable,size])	
		}
		if (length(sc1$stable) !=0) {
			newirot <- rbind(newirot, rbox1$irot[sc1$stable,-size])
			newmin <- c(newmin,pmin(rbox1$min[sc1$stable], rbox2b$min[sc1$stable]))
			newmax <- c(newmax,pmax(rbox1$max[sc1$stable], rbox2b$max[sc1$stable]))
			
		}
		if (length(sc2$stable) !=0) {
			newirot <- rbind(newirot,rbox2$irot[sc2$stable,-size])
			newmin <- c(newmin,pmin(rbox2$min[sc2$stable], rbox1b$min[sc2$stable] ))
			newmax <- c(newmax,pmax(rbox2$max[sc2$stable], rbox1b$max[sc2$stable] ))			
		}
		if (length(stable) !=0) {
			newirot <- rbind(newirot, newconstraints[stable ,-size])  # I should remove constraints I have in rbox1 and rbox2
			newmin <- c(newmin, stablemin)
			newmax <- c(newmax, -newconstraints[stable,size])			
		}
				
		awful <- setdiff(1:(size-1),c(sc1$stable, sc1$bistable))
		if (length(awful)!=0) {
			newirot <- rbind(newirot,rbox1$irot[awful,-size])
			newmin <- c(newmin,pmin(rbox1$min[awful], rbox2b$min[awful]))
			newmax <- c(newmax,pmax(rbox1$max[awful], rbox2b$max[awful]))			
		}
				
		newirot <- newirot[-1,]
		newmin <- newmin[-1]
		newmax <- newmax[-1]
		
		# we convert newirot to double since qr does not work for bigq
		newirot.double <- matrix(as.numeric(newirot), nrow=nrow(newirot), ncol=ncol(newirot)) 
		rows <- qr(t(newirot.double))$pivot[1:(size-1)]				
		result <- new(newmin[rows], newmax[rows], irot=newirot[rows,], expand=TRUE)
		
		if (DEBUG) {
			print("result")
			print(result)
		}			
		result
	}
	
	intersection <- function(rbox1,rbox2) {
		if (is.empty(rbox1)) return(rbox1)
		if (is.empty(rbox2)) return(rbox2)
		rbox1bis <- newrot(rbox2, rbox1$irot)
		new( pmax(rbox1$min, rbox1bis$min), pmin(rbox1$max, rbox1bis$max), irot=rbox1$irot)
	}	
	
	## Widening/Narrowing
	
	widening <- function(rbox1,rbox2,delay = 0) {
		if (is.empty(rbox1)) return(rbox2)
		if (delay>0) return(union.inverse(rbox2,rbox1))		## the order in union is important
		rbox1bis <- newrot(rbox2, rbox1$irot)
		newmin <- ifelse.bigqi(rbox1bis$min < rbox1$min, -Inf, rbox1$min)
		newmax <- ifelse.bigqi(rbox1bis$max > rbox1$max, +Inf, rbox1$max)
		new (newmin, newmax, irot=rbox1$irot)
	}
	
	
	narrowing <- function(rbox1,rbox2,delay = 0) {
		if (is.empty(rbox1)) return(rbox1)			
		rbox2bis <- newrot(rbox2, rbox1$irot)
		newmin <- ifelse.bigqi(rbox1$min == -Inf, rbox2bis$min, pmin(rbox1$min, rbox2bis$min))
		newmax <- ifelse.bigqi(rbox1$max == +Inf, rbox2bis$max, pmax(rbox1$max, rbox2bis$max))
		new (newmin, newmax, irot=rbox1$irot)
	}
	
	## Output
	
	as.character <- function(rbox,vars=paste("v",1:box$size,sep="")) {
		# This is more complex than with parallelotope without gmp since apply.gmp
		# wants the result to be a gmp	    
		indexes <- matrix(1:(length(vars)+1), nrow=1)
		vars.original <- apply(indexes,2,function (i) linear.form.display(t(rbox$irot[i,]), vars))
		vars.original <- vars.original[-length(vars.original)]				
		boxdom$as.character(tobox(rbox), vars.original)
	}
	
	test <- function() {
		pf <- full(2)
		pe <- empty(2)
		if (! identical(is.subseteq(pe,pf),TRUE)) warning("empty should be a subset of full") else print("test 1 passed")
		if (! identical(is.subset(pe,pf),TRUE)) warning("empty should be a proper subset of full") else print("test 2 passed")
		p1 <- new(c(-1,-1),c(1,1), expand=TRUE)
		p2 <- new(c(-1,-1),c(1,1), irot=matrix(c(1,1,1,-1),ncol=2, byrow=2), expand=TRUE)
		if (! identical(is.subseteq(p2,p1),TRUE)) warning("p2 should be a subset of p1") else print("test 3 passed")
		if (! identical(is.subset(p2,p1),TRUE)) warning("p2 should be a proper subset of p1") else print("test 4 passed")
		if (! identical(is.equal(p1,p1),TRUE)) warning("equality should be reflexive") else print("test 5 passed")
		p16 <- new(c(-1,-1),c(Inf,1), irot=matrix(c(1,1,1,-1),ncol=2, byrow=2), expand=TRUE)
		#if (! identical(argmin(p16,c(0,1,0)),c(0,-1))) warning("argmin") else print("test 6 passed")
		#if (! identical(argmin(p16,c(1,0,0)),c(-1,0))) warning("argmin") else print("test 7 passed")
		#if (! identical(argmax(p16,c(1,0,0)),c(Inf,Inf))) warning("argmax") else print("test 8 passed")
		p3 <- newrot(p2,diag(3))
		if (! identical(is.equal(p1,p3),TRUE)) warning("rotation wrong")  else print("test 9 passed")
		p4 <- new(c(0,-1),c(2,1), irot=matrix(c(1,-1,0,1),ncol=2,byrow=TRUE), expand=TRUE)
		if (! identical(is.equal(p4, assign(p1,1,c(1,1,1))),TRUE)) warning("error in invertible assignment") else print("test 10 passed")
		p14 <- new(c(1,-1),c(1,0), irot=matrix(c(1,0,-1,1),ncol=2,byrow=TRUE),expand=TRUE)
		p15 <- new(c(2,-2),c(2,-1), irot=matrix(c(1,0,-1,1),ncol=2,byrow=TRUE),expand=TRUE)
		if (! identical(is.equal(p15, assign(p14,1,c(1,0,1))),TRUE)) warning("error in invertible assignment 2") else print("test 11 passed")	
		p5 <- new(c(2,-1),c(2,1), irot=matrix(c(1,-1,0,1),ncol=2,byrow=TRUE), expand=TRUE)
		if (! identical(is.equal(p5, assign(p1,1,c(0,1,2))),TRUE)) warning("error in non-invertible assignment") else print("test 12 passed")
		p13 <- new(c(0,-Inf),c(0,Inf), irot=matrix(c(-1,1,0,1),ncol=2,byrow=TRUE), expand=TRUE)		
		p12 <- new(c(-Inf,0),c(Inf,0), irot=matrix(c(1,-1,0,1),ncol=2,byrow=TRUE), expand=TRUE)					
		if (! is.equal(p13, assign(p12,2,c(1,0,0)))) warning("error in non-invertible assignment 2") else print("test 13 passed")
		p6 <- new(c(-Inf,-1),c(Inf,1), expand=TRUE)
		if (! is.equal(p6,assign.unknown(p1,1))) warning("error in non-deterministic assignment") else print("test 14 passed")
		p7 <- new(c(-Inf,-1), c(Inf,1), expand=TRUE)
		if (! is.equal(p7,assign.unknown(p2,1))) warning("error in non-deterministic assignment 2") else print("test 15 passed")
		pd1 <- new(c(0,0),c(1,1),irot=matrix(c(2,1,2,-1),ncol=2, byrow=TRUE), expand=TRUE)
		pd2 <- new(c(-Inf,-1),c(Inf,1),irot=matrix(c(2,1,0,-2),ncol=2, byrow=TRUE), expand=TRUE)
		if (! is.equal(pd2,assign.unknown(pd1,1))) warning("error in non-deterministic assignment 3") else print("test 16 passed")
		pd3 <- new(c(-Inf,0),c(Inf,2),irot=matrix(c(2,1,4,0),ncol=2, byrow=TRUE), expand=TRUE)
		if (! is.equal(pd3,assign.unknown(pd1,2))) warning("error in non-deterministic assignment 4") else print("test 17 passed")
		p13 <- new(c(10,-1),c(10,1),irot=matrix(c(1,0,1,1),ncol=2,byrow=TRUE),expand=TRUE)
		p14 <- new(c(-Inf,-11),c(+Inf,-9),expand=TRUE)
		if (! is.equal(p14,assign.unknown(p13,1))) warning("error in non-deterministic assignment 5") else print("test 18 passed")
		p8 <- new(c(-1,-1),c(0,0),  irot=matrix(c(1,1,1,-1),ncol=2,byrow=TRUE), expand=TRUE) 
		if (! is.equal(p8,select(p2,c(2,0,1)))) warning("error in select bounded")	 else print("test 19 passed")
		p9 <- new(c(-1,-Inf),c(1,Inf),irot=matrix(c(1,1,1,-1),ncol=2, byrow=2), expand=TRUE)
		p10 <- new(c(-1,-Inf),c(1,-2),irot=matrix(c(1,1,1,0),ncol=2, byrow=2), expand=TRUE)
		if (! is.equal(p10,select(p9,c(1,0,2)))) warning("error in select unbounded") else print("test 20 passed")
		pa <- new(c(0,0), c(+Inf,+Inf), expand=TRUE)
		pb <- new(c(0,-Inf),c(Inf,0),irot=matrix(c(1,0,1,-1),ncol=2, byrow=TRUE), expand=TRUE)
		if (! is.equal(pb, select(pa,c(1,-1,0)))) warning("error in select unbounded 2") else print("test 21 passed")
		pa <- new(c(0,0), c(0,+Inf), expand=TRUE)
		pb <- new(c(0,-Inf),c(0,0),irot=matrix(c(1,0,1,-1),ncol=2, byrow=TRUE), expand=TRUE)
		if (! is.equal(pb, select(pa,c(1,-1,0)))) warning("error in select unbounded 3") else print("test 22 passed")
		p11 <- new(c(10,-10),c(10,-10),expand=TRUE)
		if (! is.equal(pe,select(p11,c(1,-1,0)))) warning("error in select on p11") else print("test 23 passed")		
		p99 <- new.simple(c(2,0),c(4,2))
		if (! is.equal( new.simple(c(-4,-1),c(4,4),c(-1,3,1,0)),union.inverse(p1,p99))) warning("error in inverse union") else print("test 24 passed")
		if (! is.equal( p1,union.inverse(p1,p1))) warning("error in inverse union 2")		
		if (! is.equal( new.simple(c(-6,-4),c(4,2),c(1,-4,-1,2)),union.inverse(p99,p2))) warning("error in inverse union 3") else print("test 25 passed")
		p98 <- new.simple(c(-4,0),c(-2,2))
		if (! is.equal( new.simple(c(-4,0),c(4,2)),union.inverse(p98,p99))) warning("error in inverse union 4")	else print("test 26 passed")	
		p96 <- new.simple(c(1,-Inf),c(1,1),c(1,0,1,-1))
		p97 <- new.simple(c(0,-Inf),c(0,0),c(1,0,0,-1))				
		if (! is.equal( new.simple(c(0,0),c(1,Inf)),union.inverse(p96,p97))) warning("error in inverse union 5") else print("test 27 passed")
		p95 <- new.simple(c(0,0),c(0,Inf))
		p94 <- new.simple(c(0,0),c(1,Inf))
		if (! is.equal( p94,union.inverse(p95,p94))) warning("error in inverse union 6") else print("test 28 passed")
		if (! is.equal( p94,union.inverse(p94,p95))) warning("error in inverse union 7") else print("test 29 passed")						
	}	
	
	setparam <- function(rbox, v) {
		rbox
	}
	
	outputform <- function(box,lf,vars=paste("v",1:box$size,sep="")) {
		min.bound <- minimize(box,lf)
		max.bound <- maximize(box,lf)		
		boxdom$as.character(list(size=1, min=c(min.bound,1), max=c(max.bound,1)), linear.form.display(lf,vars))
	}
	
	## Declaratation of public functions
	
	list ( new=new, full = full, empty = empty, zero=zero, frombox=frombox, getsize=getsize, newrot=newrot,
			is.subseteq = is.subseteq, is.subset = is.subset, is.equal = is.equal, is.empty = is.empty,
			argmin = argmin, argmax = argmax, minimize = minimize, maximize = maximize,
			union = union.weak, intersection = intersection, widening = widening, narrowing=narrowing,
			assign = assign, assign.unknown = assign.unknown, select = select, select.diff = select.diff,
			as.character=as.character, test=test, new.simple =  new.simple, union.inverse=union.inverse,
			setparam=setparam, outputform=outputform )
}

#pg <- generalptopes.domain.gmp()
#pg$test()