## S3 class bigqi
## Wrapper for bigq which adds the values +Inf and -Inf 
## and correctly manipulates NA
## (and NaN possibly one day)

## Copyright 2008-2012 Gianluca Amato, Francesca Scozzari

## This file is part of RANDOM: R-based Analyzer for Numerical DOMains
##
## RANDOM is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## RANDOM is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with RANDOM.  If not, see <http://www.gnu.org/licenses/>.

# Redefine as.bigq as an S3 method, since we define a new as.bigq.bigqi method.

as.bigq.default <- function(x,...) {
   if (length(x)==0) 
     rep(gmp::as.bigq(0),0)
   else
      gmp::as.bigq(x,...)
}
as.bigq <- function(x,...) UseMethod("as.bigq")

# A bigqi is implemented as a list with class bigqi and two components:
#  - bigq is a bigq (which is a vector of rational numbers in the library gmp)
#  - inf is a vector of integers corresponding to the elements in bigq. The value 
#    0 means that the value in bigq is valid, +1 denotes a +Inf and -1 denotes -Inf.
#    If inf!=0, the corresponding bigq should be 0. 
# NA value are represented with an NA in both the bigq and the inf components

# The representation is wrong since the attributes names should name the position
# of the numbers in the bigqi vector, not the names of the internal component. The
# right approach is to use a bigq with an additional inf attribute.

# bigqi.make: internal function for generating bigqi objects.
# Params: the two compononents bigq and inf
# Returns: a bigqi object
bigqi.make <- function(bigq, inf) {  
  bigqi <- list(bigq=as.vector(bigq), inf=inf)
  class(bigqi) <- "bigqi"
  return(bigqi)
}

# bigqi.check: check consistency of a bigqi to its definition above
# Params: qi: the bigqi to check
# Returns: whether qi is consistent to bigqi definition
bigqi.check <- function(qi) {
  check1 <- rep(TRUE, length(qi$bigq))
  inf.indexes <- ifelse(is.na(qi$inf),FALSE,qi$inf != 0)
  if (any(inf.indexes))
    check1[inf.indexes] <- qi$bigq[inf.indexes] == 0
  na1.indexes <- is.na(qi$inf)
  na2.indexes <- is.na(qi$bigq)
  check2 <- na1.indexes == na2.indexes
  check3 <- nrow(qi$bigq)==0
  return(check1 & check2 & check3)
}

# bigqi: create a bigqi vector
# Params: length: desired length
# Returns: a vector of bigqi filled with zero of length n
bigqi <- function(length=0) {
  bigq <- rep(as.bigq(0),length)
  inf <- integer(length)
  bigqi.make(bigq,inf)
}

# as.bigqi: converts the input into a bigqi
# Params: x: object to be coerced
as.bigqi <- function (x,y=1) {
  if(!missing(y)) return(as.bigqi(x)/as.bigqi(y))
  if (length(x)==0) return(bigqi(0))
  if (class(x)=="bigqi") return(x) 
  if (class(x)!="bigq") {
    bigq <- as.bigq(x)
    inf <- as.integer(is.infinite(x)) *  sign(x)
    bigq[[inf!=0 & !is.na(inf)]] <- 0 
  } else {   
    bigq <- x
    inf <- integer(length(x))
  }
  bigqi.make(bigq,inf)
}


# is.bigqi: tests whether the input is a bigqi
# Params: x: object to be tested
is.bigqi <- function (x) {
  return (class(x)=="bigqi")
}

# length.bigqi: returns the length of a bigqi value
length.bigqi <- function (qi) { 
  length(qi$bigq) 
}

# as.character.bigqi: converts a biqi into a string
# Params: a bigqi object
as.character.bigqi <- function(qi) {
  as.character(  ## this is needed in case qi has length 0 so that the result
                 ## is a character instead of a logical vector
     ifelse(is.na(qi$inf) | qi$inf==0, as.character(qi$bigq), ifelse(qi$inf==1, "Inf", "-Inf")) 
  )
}

# as.bigq.bigqi: converts a bigqi into a bigq
# Params: a bigqi object
# Returns: a bigq object
# Trows: a warning if the bigqi objects contains infinite values
as.bigq.bigqi <- function(qi) {
  if (any(qi$inf!=0)) warning("cannot convert a bigqi with infinite values")
  return(qi$bigq)
} 

# print.bigqi: prints a bigqi. It essentially print the as.character version of
# its input, but if it is empty it prints "bigqi(0)"
print.bigqi <- function(qi,...) {
  if (length(qi$bigq)>0)
    print(as.character(qi))
  else
    cat("bigqi(0)\n")
}

# [.bigqi: extracts part of a bigqi
# Params: qi: a bigqi object
#         i: index specifying element to extract
# Returns: the extractes subvector
"[.bigqi" <- function(qi,i) {
  bigq <- qi$bigq[[i]]
  inf <- qi$inf[i]
  return(bigqi.make(bigq,inf))
}

# [<-.bigqi: replace part of a bigqi
# Params: qi: a bigqi object
#         i: index specifying elements to replace
#         value: new objects. Note that we cannot change the name of this parameter.
# Returns: qi with elements indexed by i replaced with elements in qj
"[<-.bigqi" <- function(qi,i,value) {
  value <- as.bigqi(value)
  qi$bigq[[i]] <- value$bigq
  qi$inf[i] <- value$inf
  return(qi)
}

# *.bigqi: multiplication between bigqi
# Params: qi1, qi2: bigqi vectors 
# Returns: the bigqi corresponding to qi1*qi2
"*.bigqi" <- function(qi1,qi2) {
  if (length(qi1)==0 || length(qi2)==0) return(bigqi(0))
  ## coerce types
  qi1 <- as.bigqi(qi1)
  qi2 <- as.bigqi(qi2)
  ## compute product
  bigq <- qi1$bigq * qi2$bigq
  inf <- qi1$inf * qi2$inf + qi1$inf * sign(qi2$bigq) + qi2$inf * sign(qi1$bigq)
  infres <- inf!=0 & !is.na(inf)
  if (any(infres)) {
    bigq[[infres]]<-0
  }  
  nares <- (qi1$bigq==0 & qi1$inf==0 & qi2$inf!=0) | (qi2$bigq==0 & qi2$inf==0 & qi1$inf!=0)
  nares <- nares & !is.na(nares)
  if (any(nares)) {  
    bigq[[nares]] <- NA
    inf[nares] <- NA 
  }
  bigqi.make(bigq,inf) 
}

# +.bigqi: sum between bigqi
# Params: qi1, qi2: bigqi vectors 
# Returns: the bigqi corresponding to qi1+qi2
"+.bigqi" <- function(qi1,qi2) {
  if (length(qi1)==0 || length(qi2)==0) return(bigqi(0))
  ## coerce types
  qi1 <- as.bigqi(qi1)
  qi2 <- as.bigqi(qi2)
  ## compute sum
  bigq <- qi1$bigq + qi2$bigq
  inf <- sign(qi1$inf + qi2$inf)
  infres <- inf!=0 & !is.na(inf)
  if (any(infres)) {
    bigq[[infres]]<-0
  }  
  nares <- qi1$inf * qi2$inf < 0 
  nares <- nares & !is.na(nares)
  if (any(nares)) {  
    bigq[[nares]] <- NA
    inf[nares] <- NA 
  }
  bigqi.make(bigq,inf) 
}

# inverse: the inverse of a bigqi
# Params: qi: bigqi vector
# Returns: the inverse of qi
inverse <- function(qi) {
  if (length(qi)==0) return(bigqi(0))
  ## coerce types
  qi <- as.bigqi(qi)
  bigq = rep(as.bigq(0),length(qi))
  normalinverse = is.na(qi$bigq) | (qi$bigq!=0)
  if (any(normalinverse)) bigq[[normalinverse]] = 1/qi$bigq[[normalinverse]]
  inf = as.integer(qi$bigq==0 & qi$inf==0)
  bigqi.make(bigq,inf)
} 

# opposite: the opposite of a bigqi
# Params: qi: bigqi vector
# Returns: the inverse of qi
opposite <- function(qi) {
  if (length(qi)==0) return(bigqi(0))
  ## coerce types
  qi <- as.bigqi(qi)
  bigqi.make(-qi$bigq, -qi$inf)
}

# Other operators... the pattern is clear

"-.bigqi" <- function(qi1,qi2=NULL) {
  if (missing(qi2))
    opposite(qi1)
  else
    qi1 + opposite(qi2)
}

"/.bigqi" <- function(qi1,qi2) {
  qi1 * inverse(qi2)
}

">.bigqi" <- function(qi1,qi2) {
  if (length(qi1)==0 | length(qi2)==0) return(logical(0))
  ## coerce types
  qi1 <- as.bigqi(qi1)
  qi2 <- as.bigqi(qi2)
  (qi1$inf > qi2$inf) | (qi1$inf == qi2$inf  &  qi1$bigq > qi2$bigq)  
}

"<.bigqi" <- function(qi1,qi2) {
  if (length(qi1)==0 | length(qi2)==0) return(logical(0))
  ## coerce types
  qi1 <- as.bigqi(qi1)
  qi2 <- as.bigqi(qi2)
  (qi1$inf < qi2$inf) | (qi1$inf == qi2$inf  &  qi1$bigq < qi2$bigq)  
}

">=.bigqi" <- function(qi1,qi2) {
  if (length(qi1)==0 | length(qi2)==0) return(logical(0))
  ## coerce types
  qi1 <- as.bigqi(qi1)
  qi2 <- as.bigqi(qi2)
  (qi1$inf > qi2$inf) | (qi1$inf == qi2$inf  &  qi1$bigq >= qi2$bigq)  
}

"<=.bigqi" <- function(qi1,qi2) {
  if (length(qi1)==0 | length(qi2)==0) return(logical(0))
  ## coerce types
  qi1 <- as.bigqi(qi1)
  qi2 <- as.bigqi(qi2)
  (qi1$inf < qi2$inf) | (qi1$inf == qi2$inf  &  qi1$bigq <= qi2$bigq)  
}

"==.bigqi" <- function(qi1,qi2) {
  if (length(qi1)==0 | length(qi2)==0) return(logical(0))
  ## coerce types
  qi1 <- as.bigqi(qi1)
  qi2 <- as.bigqi(qi2)
  (qi1$inf == qi2$inf) & (qi1$bigq == qi2$bigq)  
}

"!=.bigqi" <- function(qi1,qi2) {
  if (length(qi1)==0 | length(qi2)==0) return(logical(0))
  ## coerce types
  qi1 <- as.bigqi(qi1)
  qi2 <- as.bigqi(qi2)
  (qi1$inf != qi2$inf) | (qi1$bigq != qi2$bigq)  
}

is.infinite.bigqi <- function(qi) { 
  qi$inf != 0
}

is.finite.bigqi <- function(qi) {
   qi$inf == 0
}

is.na.bigqi <- function(qi) {
   is.na(qi$bigq)
}

is.nan.bigqi <- function(qi) {
   # since we de not distinguish between NA and NaN, this is equivalent
   # to is.na
   is.na(qi$bigq)  
}

# rep.bigqi needs to manually support length.out since rep.bigq does not.
# we should generate a warning when reps is not an integer
rep.bigqi <- function (qi,...,length.out=NULL) {
  if (missing(length.out))
    bigqi.make(rep(qi$bigq,...), rep(qi$inf,...))
  else {
    reps = as.integer(length.out / length(qi))
    bigqi.make(rep(qi$bigq,reps),rep(qi$inf,reps))
  }
}

## to be rewritten...
c.bigqi <- function(...) {
  a <- list(...)
  i <- 1
  bigq <- rep(as.bigq(0),0)
  inf <- integer(0)
  l <- length(a)
  while (i <= l) {
    a[[i]] <- as.bigqi(a[[i]])
    bigq <- c(bigq,a[[i]][["bigq"]])
    inf <- c(inf,a[[i]][["inf"]])
    i <- i+1
  }
  bigqi.make(bigq, inf)
}

## The option na.rm is not supported
min.bigqi <- function(..., na.rm=FALSE) {
  param = c(...)
  if(any(param == -Inf))
    return(as.bigqi(-Inf))
  else if(all(param == +Inf))
    return(as.bigqi(+Inf))
  else { 
    newparam = param[param!=+Inf]
    return(as.bigqi(min(newparam$bigq)))
  }
}

## The option na.rm is not supported
max.bigqi <- function(..., na.rm=FALSE) {
  param = c(...)
  if(any(param == +Inf))
    return(as.bigqi(+Inf))
  else if(all(param == -Inf))
    return(as.bigqi(-Inf))
  else { 
    newparam = param[param!=-Inf]
    return(as.bigqi(max(newparam$bigq)))
  }
}

## This is to override standard ifelse which do not work well with 
## raw objects
ifelse.bigqi <- function(test,yes,no) {
    ans <- bigqi(length(test))
    nas <- is.na(test)
    if (any(test[!nas])) 
        ans[test & !nas] <- rep(yes, length.out = length(ans))[test & 
            !nas]
    if (any(!test[!nas])) 
        ans[!test & !nas] <- rep(no, length.out = length(ans))[!test & 
            !nas]
    ans[nas] <- NA
    ans
}

## The function pmin should be removed when
## we move to the new internal representation.
pmin.bigqi <- function(v1,v2) {
	v1 <- as.bigqi(v1)
	v2 <- as.bigqi(v2)
    result <- bigqi(0)
	for (i in 1:length(v1)) {
		result <- c(result,
			if (is.na(v1[i]) || is.na(v2[i]))
				as.bigqi(NA)
			else if (v1[i]<=v2[i]) 
				v1[i] else v2[i]
		)
	} 
		
	return(result)
}

## The function pmax should be removed when
## we move to the new internal representation.
pmax.bigqi <- function(v1,v2) {
   v1 <- as.bigqi(v1)
   v2 <- as.bigqi(v2)
   result <- bigqi(0)
   for (i in 1:length(v1)) {	   
	   result <- c(result,
		if (is.na(v1[i]) || is.na(v2[i]))
	  	    as.bigqi(NA)
		else  if (v1[i]>=v2[i]) 
			v1[i] else v2[i])
   } 
   return(result)
}

# This does not support na.rm
sum.bigqi <- function(...,na.rm = FALSE) {
   param = c(...)
   if (any(param$inf==1) && any(param$inf==-1)) 
     as.bigqi(NA)
   else if (any(param$inf==1))
     as.bigqi(+Inf)
   else if (any(param$inf==-1)) 
     as.bigqi(-Inf)
   else
     as.bigqi(sum(param$bigq))
   
}
