## RANDOM
## This is a re-implementation of rbind.fill of the library plyr, in such
## a way that it works for lists. We also reimplement the empty function.

## Copyright Hadley Wickham <h.wickham@gmail.com>
## Copyright 2008-2011 Gianluca Amato, Francesca Scozzari

## This file is part of RANDOM: R-based Analyzer for Numerical DOMains
##
## RANDOM is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## RANDOM is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty ofa
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with RANDOM.  If not, see <http://www.gnu.org/licenses/>.

empty <- function(l) { 
	is.null(l) || length(l)==0 || (is.data.frame(l) && .row_names_info(l) == 0)
}

rbind.fill <- function(...) {
	dfs <- list(...)
	if (length(dfs) == 0) return()
	if (is.list(dfs[[1]]) && !is.data.frame(dfs[[1]])) {
		dfs <- dfs[[1]]
	}
	dfs <- Filter(Negate(empty), dfs)
	
	if (length(dfs) == 0) return()
	if (length(dfs) == 1) return(dfs[[1]])
	
	# Calculate rows in output
	# Using .row_names_info directly is about 6 times faster than using nrow	
	rows <- unlist(lapply(dfs, function(df) if (is.data.frame(df)) .row_names_info(df,2L) else length(df[[1]])))  # MODIFIED!!
	nrows <- sum(rows)
	
	# Generate output template
	output <- output_template(dfs, nrows)
	
	# Compute start and end positions for each data frame
	pos <- matrix(cumsum(rbind(1, rows - 1)), ncol = 2, byrow = TRUE)
	
	# Copy inputs into output
	for(i in seq_along(rows)) {
		rng <- pos[i, 1]:pos[i, 2]
		df <- dfs[[i]]
		
		for(var in names(df)) {
			if (!is.matrix(output[[var]])) {
				output[[var]][rng] <- df[[var]]
			} else {
				output[[var]][rng, ] <- df[[var]]
			}
		}
	}
	
	quickdf(output)
}

output_template <- function(dfs, nrows) {
	vars <- unique(unlist(lapply(dfs, base::names))) # ~ 125,000/s
	output <- vector("list", length(vars))
	names(output) <- vars
	
	seen <- rep(FALSE, length(output))
	names(seen) <- vars
	
	is_array <- seen
	is_matrix <- seen
	is_factor <- seen
	
	for(df in dfs) {
		matching <- intersect(names(df), vars[!seen])
		for(var in matching) {
			value <- df[[var]]
			
			if (is.vector(value) && is.atomic(value)) {
				output[[var]] <- rep(NA, nrows)
			} else if (is.factor(value)) {
				output[[var]] <- factor(rep(NA, nrows))
				is_factor[var] <- TRUE
			} else if (is.matrix(value)) {
				is_matrix[var] <- TRUE
			} else if (is.array(value)) {
				is_array[var] <- TRUE
			} else if (inherits(value, "POSIXt")) {
				output[[var]] <- as.POSIXct(rep(NA, nrows))
				attr(output[[var]], "tzone") <- attr(value, "tzone")
			} else if (is.list(value)) {
				output[[var]] <- vector("list", nrows)
			} else {
				output[[var]] <- rep(NA, nrows)
				class(output[[var]]) <- class(value)
				attributes(output[[var]]) <- attributes(value)
			}
		}
		
		seen[matching] <- TRUE
		if (all(seen)) break # Quit as soon as all done
	}
	
	# Set up factors
	for(var in vars[is_factor]) {
		all <- unique(lapply(dfs, function(df) levels(df[[var]])))
		output[[var]] <- factor(output[[var]], levels = unique(unlist(all)),
				exclude = NULL)
	}
	
	# Set up matrices
	for(var in vars[is_matrix]) {
		width <- unique(unlist(lapply(dfs, function(df) ncol(df[[var]]))))
		if (length(width) > 1)
			stop("Matrix variable ", var, " has inconsistent widths")
		
		vec <- rep(NA, nrows * width)
		output[[var]] <- array(vec, c(nrows, width))
	}
	
	# Set up arrays
	for (var in vars[is_array]) {
		dims <- unique(unlist(lapply(dfs, function(df) dims(df[[var]]))))
		if (any(dims) > 1) {
			stop("rbind.fill can only work with 1d arrays")
		}
		
		output[[var]] <- rep(NA, nrows)
	}
	
	output
}
