## Simplifiers for non-integer matrices. It implements OSCA and 
## an house-made simplification based on GCD.
##
## From the paper:
##   Orthogonal Simple Component Analysis
##   by Karim Anaya-Izquierdo, Frank Critchley, Karen Vines
##   The Open University
##

## Copyright 2008-2012 Gianluca Amato

## This file is part of RANDOM: R-based Analyzer for Numerical DOMains
##
## RANDOM is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## RANDOM is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with RANDOM.  If not, see <http://www.gnu.org/licenses/>.

##
## select.non.singular(Z): given an m*n matrix with m>=n>=0, determines
##   a permutation of the columns such that the top n*n matrix is not singular.
##   Gives an error if such a permutation does not exists.
##
select.non.singular = function(Z) {
  qr = qr(t(Z))
  if (qr$rank != ncol(Z))
    stop("Error: matrix Z has not maximun rank")
  else
    qr$pivot
}

##
## cofactors(Z): given an n*n matrix (n>=0), determines the matrix of
##   cofactors
##
cofactors = function(Z) {
  n = nrow(Z)
  if (n!= ncol(Z)) stop("Matrix Z is not square")
  res = matrix(nrow=n,ncol=n)
  for (i in seq(1,by=1,length.out=n))
    for (j in seq(1,by=1,length.out=n))
      res[i,j] = ((-1)^(i+j))*round(det(Z[-i,-j,drop=FALSE]))
  ## Due to its implementation, the function det() may return floating point values
  ## even if its argument is integer.  This is the reason we also used round()
  return(res)
}

##
## orthogonal.space(Z): given an m*n integer matrix with m>=n>=0, determines a
##   matrix whose column is a base of the orhogonal space given by the columns
##   in Z
##
orthogonal.space = function(Z) {
  k = nrow(Z)
  r = ncol(Z)
  perm = select.non.singular(Z)
  Za = Z[perm[seq(1,by=1,length.out=r)],,drop=FALSE]
  Zb = Z[perm[seq(r+1,by=1,length.out=k-r)],,drop=FALSE]
  O = rbind( -cofactors(Za)%*% t(Zb), round(det(Za)) * diag(k-r))
  ## Due to its implementation, the function det() may return floating point values
  ## even if its argument is integer.  This is the reason we also used round()
  return(O[inverse.permutation(perm),,drop=FALSE])
}

##
## accuracy(v,w): given two vectors in R^n, computes the accuracy, i.e.
##   the cosine of their angle
##
accuracy = function(v,w) {
  acc=abs(sum(v * w))/sqrt(sum(v^2)*sum(w^2))
  if(is.nan(acc))
	  0
  else 
	  acc
}

##
## inverse.permutation(x): computes the inverse permutation
##
inverse.permutation = function(x) {
  l = integer(length(x))
  l[x] = 1:length(x)
  return(l)
}

##
## find.best.approximation: in the space generated by columns in O,
##   search for the integer vector which has minimum complexity, among
##   those whose accuracy w.r.t. q is bigger than theta. Nmax and Imax
##   are search parameter of the paper.
##   
##
find.best.approximation = function(O, q, theta, Nmax=9, Imax=20000) {
  k = nrow(O)         ## dimension of the vector space
  r = k-ncol(O)+1     ## dimension to fill

  ## project q in the orthogonal space  
  ystar = solve(t(O) %*% O) %*% t(O) %*% q

  ## record negative components in ystar 
  ystar.sign = sign(ystar)

  ## record ordering of components in ystar
  ystar.ordering = order(abs(ystar),decreasing=TRUE)
  ystar.revordering = inverse.permutation(ystar.ordering)
  
  ## initalize results
  
  max.accuracy = -Inf
  min.complexity = +Inf
  best.vector = NA

  for (N in 1:Nmax) {

    ## computes number of step required for the search
    if (k==r)
      IN=1
    else
      IN = sum(sapply( 1:(min(N,k-r)+1) ,
                   function(i)  choose(k-r-1,i-1)*choose(N+1,i) ))

    if (IN<=Imax) {
      v = rep(N,k-r+1)
      j = k-r+1
      while (TRUE) {
        if (j==k-r+1)  {
          IN = IN - 1
          w = O %*% (v[ystar.revordering]*ystar.sign)
          #gcdW=gcd(w)
		  #if(gcdW!=0) w = w/gcdW
          a = accuracy(w,q)
          if (a >= theta)
            if (max(abs(w))<min.complexity) {
              min.complexity = max(abs(w))
              max.accuracy = a
              best.vector = w
              best.v = v
            } else if (max(abs(w))==min.complexity && a > max.accuracy) {
              max.accuracy=a
              best.vector = w
              best.v = v
            }
          ## move cursor to the last non-zero position
          while(v[j]==0 && j>1)  j=j-1
          ## decrement data under cursor
          if (j!=1) v[j] = v[j]-1 else break;                   
        } else {
          ## put all next coordinates to same coordinate then cursor
          v[(j+1):(k-r+1)] = v[j]
          j = k-r+1
        }
      }
      if (IN!=0) stop("Error: wrong number of steps")
    } else {
      stop("IN>Imax not implemented yet")
    }
  }
  return (list(vec=best.vector, accuracy=max.accuracy, complexity = min.complexity))
}

##
## OSCA simplification algorithm
##
## OSCAsimplify(Q,Nmax,Imax): given an n*n matrix, determines an integer n*n
## matrix whose columns are pairwise orthogonal. If orthogonal is false,
## the matrix is not orhogonal.
##
OSCAsimplify = function(Q, theta=pi/4, Nmax=9, Imax=20000, orthogonal=TRUE, reverse=TRUE) {
  if (nrow(Q) != ncol(Q)) stop ("Matrix Q is not square")
  if (nrow(Q)==0 || ncol(Q)==0) {
	  Z <- Q
	  attr(Z,"accuracy") <- 1
	  return(Z)
  }
  
  k <- nrow(Q)
  Z <- matrix(nrow=k, ncol=0) 
  accuracy = numeric(0)
  if (reverse) order.columns = k:1 else order.columns = 1:k
  for (r in order.columns) {
    if (orthogonal) 
      O <- orthogonal.space(Z)
    else
      O <- diag(k)
    best.approx = find.best.approximation(O,Q[,r],theta,Nmax,Imax)
    if (best.approx$accuracy == -Inf) {		
	  # The next message is hard-coded in trace-analyzers.R and should be changed in sync		
      stop("You got a corner case of OSCA which has not been implemented yet")
    }
    Z <- cbind(Z,best.approx$vec)
    accuracy <- c(accuracy,best.approx$accuracy)
  }   
  Z <- Z[,order.columns, drop=FALSE]
  accuracy <- accuracy[order.columns]
  rownames(Z) <- rownames(Q)
  colnames(Z) <- colnames(Q)
  attr(Z,"accuracy") <- accuracy
  return(Z)
}
			   
##
## GCD simplification algorithm
##
## gcdSimplification(matrix, debug.simplify=FALSE, zero.threshold.simplification=1/10): 
## given an n*n matrix, determines an integer n*n matrix maintaining the ratio 
## between elements in a same row. If debug.simplify is TRUE, it outputs some messages to
## help debugging.
##
GCDsimplify = function (M, zero.threshold.simplification=1/10) { 	
	if (nrow(M) != ncol(M)) stop ("Matrix Q is not square")
	if (nrow(M)==0 || ncol(M)==0) return(Q)
	# A trick for the problem that apply transform row/column matrices in vectors	
	R <- apply(M, 2, function(r) c(0,round(r/gcd(r, zero.threshold = zero.threshold.simplification))))
	R[-1,,drop=FALSE]
}
