## RANDOM
## This file contains the trace analyzers and matrix simplifiers used in RANDOM.

## Copyright 2008-2012 Gianluca Amato, Francesca Scozzari

## This file is part of RANDOM: R-based Analyzer for Numerical DOMains
##
## RANDOM is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## RANDOM is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with RANDOM.  If not, see <http://www.gnu.org/licenses/>.

traceAnalyzerCondition <- function(message) {
	cond <- simpleCondition(message)
	class(cond) <- c("traceAnalyzerCondition",class(cond))
	cond
}

# This is the class for statistical analysis. There are three attributes:
# call is the call which it is referring
# matrix is the default transformation matrix
# aggregates is the per program point transformation matrix
# the method simplifyOSCA simplifies a matrix using the OSCA algorithm
# the method simplifyGCD uses the GCD simplification


StatisticalTraceAnalysis <- proto(
	new = function(., call, matrix, aggregates=NULL) {
		.$proto(call = call, matrix = matrix, aggregates = aggregates)
	},	
	simplifyOSCA = function(., forceint = c("yes","no","experimental"), ...) {		
		getOSCA <- function(matrix, forceint, ...) {			
			if (attr(matrix,"inverse")) 
				matrix <- t(matrix)
			tryCatch (
			  switch (forceint,
				  	  yes = { osca <- t(OSCAsimplify(matrix,...)); attr(osca,"inverse") <- TRUE },
					  no = { osca <- OSCAsimplify(matrix,...); attr(osca,"inverse") <- FALSE },
					  experimental = { osca <- OSCAsimplify(t(matrix),...); attr(osca,"inverse") <- TRUE }
			  ),
			  silent = TRUE,
			  error = function(m) 
			  	if (conditionMessage(m)=="You got a corner case of OSCA which has not been implemented yet")
	 				signalCondition(traceAnalyzerCondition(
					 paste(conditionMessage(m),"Please try lowering the threshold or using GCD simplification.")))
				else 
					stop(m)
	  		)
			osca
		}		
		forceint <- match.arg(forceint)
		matrix2 <- getOSCA(.$matrix, forceint, ...)
		aggregates2 <- lapply(.$aggregates, function(x) 
					if (is.null(x)) NULL else tryCatch( getOSCA(x, forceint, ...), error = function(e) NULL )
		)
		StatisticalTraceAnalysis$new(call = .$call, matrix = matrix2, aggregates = aggregates2 ) 
	},
	simplifyGCD = function(., ...) {
		getGCDSimplification <- function(matrix, ...) {			
			if (attr(matrix,"inverse")) 
				matrix <- t(matrix)
			experSimpl <- t(GCDsimplify(matrix, ...)); 
			attr(experSimpl,"inverse") <- TRUE
			experSimpl		
		}		
		matrix2 <- getGCDSimplification(.$matrix, ...)
		aggregates2 <- lapply(.$aggregates, function(x) 
					if (is.null(x)) NULL else tryCatch( getGCDSimplification(x, ...), error = function(e) NULL )
		)
		StatisticalTraceAnalysis$new(call = .$call, matrix = matrix2, aggregates = aggregates2 ) 
	}
)

# this is the factory for statistical trace analysis
statisticalTraceAnalyzer <- proto(
		fromTrace = function(., f, programTrace, per.pp = FALSE, ...) {
			data <- programTrace$getCleanData()		
			if (empty(data)) {
				.$identity(programTrace$call, setdiff(programTrace$getVars(),".point"))
			} else {
				matrix <- f(data[-1], ...)
				if (per.pp) 
					aggregates <- lapply (programTrace$getCleanAggregatedData(), function(x) tryCatch(f(x,...), error=function(e) NULL))
				else
					aggregates <- list()
				StatisticalTraceAnalysis$new(call = programTrace$call, matrix = matrix, aggregates = aggregates )	
			}			
		},
		
		PCAfromTrace = function(., programTrace, ...) {
			getpca <- function(data) {
				pca <- prcomp(data)$rotation
				attr(pca,"inverse") <- FALSE
				pca
			}
			
			.$fromTrace(getpca, programTrace, ...)					
		},
		
		ICAfromTrace = function(., programTrace, zero.threshold=0.01, debug.trace.analyzer=FALSE,...) {		
			getica <- function(data) {
				# pca <- prcomp(data,retx=TRUE)
				pca <- prcomp(data)
				n=ncol(pca$rotation)
				sdev.zeroes <- sum(pca$sdev < zero.threshold)
				n.component=n-sdev.zeroes
				
				# compute fastICA several times decrementing n.comp when raising an error
				# does not work when method="C", since it does not rise the error condition
				recomputeICA=TRUE
				while(n.component>0 && recomputeICA) {
					tryCatch(  
							{
								# result <- fastICA(pca$x[,1:n.component], n.comp = n.component)
								# result <- fastICA(data, n.comp = n.component, method="C",w.init=diag(n.component),maxit=10000,tol=0.000001)
								
								result <- fastICA(data, n.comp = n.component, method="C", 
										w.init=pca$rotation[(n-n.component+1):n,(n-n.component+1):n],maxit=10000,tol=0.000001)
								recomputeICA=FALSE
							} ,
							error=function(e) NULL, 
							finally= {if(recomputeICA) {n.component <- n.component-1}}
					)
				}
				if(n.component != 0) { 
					ica <- t(result$K %*% result$W)
					ica <- rbind(ica,t(pca$rotation)[seq(n.component+1,length.out=sdev.zeroes),])
					attr(ica,"inverse") <- TRUE
				} else {
					ica <- t(pca$rotation)
					attr(ica,"inverse") <- TRUE
				}

				# debugging comments 
				if(debug.trace.analyzer) {
					print("PCA sdev")
					print(pca$sdev)
				}

				colnames(ica) <- colnames(data)
				rownames(ica) <- paste("PC",1:ncol(data),sep=""); 
				ica
			}			
			.$fromTrace(getica, programTrace, ...)
		},
				
		identity = function(., call, vars) {
			if(length(vars)==0) 
				m <- matrix(nrow=0, ncol=0)
			else {
				m <- diag(length(vars))
				colnames(m) <- vars
				rownames(m) <- paste("PC",1:(length(vars)),sep="")
			}			
			attr(m,"inverse") <- TRUE				
			StatisticalTraceAnalysis$new(call = call, matrix = m, aggregates=NULL)
		}
)
