## RANDOM
## This is the program tracer.

## Copyright 2008-2012 Gianluca Amato, Francesca Scozzari

## This file is part of RANDOM: R-based Analyzer for Numerical DOMains
##
## RANDOM is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.
##
## RANDOM is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty ofa
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with RANDOM.  If not, see <http://www.gnu.org/licenses/>.

source("libraries/rbind.fill.R")
source("annotate.R")

programTracer <- proto(
		# fromFunction builds a ProgramTrace by executing a program. 
		# if params is NULL, fun should be a call of the kind functioname(parameters)
		#   otherwise fun should be a function and params a list of parameters as inputs for do.call
		#   otherwise fun should be a function and params a list of list of parameters		
		# vars is the set of variables we want to trace: NULL if we want to trace all the variables		
		# noscalar decides what to do with nonscalar variables: yes preserves the variable, 
		#   remove removes them and na converts the to NA
		# tag determines what to tag as a program point: if yes, all program points are tagged, if
		#   onlywhile, only the first instruction in a while is tagged, if no it is assumed the function
		#   is already tagged.
		# maxsteps is the maximum numbers of steps to trace (maxsteps=1000L)
		
		fromFunction = function(.,  fun, params=NULL, ..., tag = c("yes","onlywhile", "no"), maxsteps=1000L, noscalar=c("yes","no","na"), vars=NULL ) {
			
			tracerCondition <- function(message) {
				cond <- simpleCondition(message)
				class(cond) <- c("tracerCondition",class(cond))
				cond
			}			
			
			# currpoint is an index used to label the program points 
			currpoint <- 0L
			
			# maxsteps is converted to an integer, otherwise identical() does not work
			maxsteps <- as.integer(maxsteps)
			
			noscalar <- match.arg(noscalar)
			tag <- match.arg(tag)
			
			# collectlist is the list used to store the trace data at each program point
			collectlist <- list()				
			
			# convert input parameters in such a way that params is always not null and 
			# is a list of input parameters
			if (is.null(params)) {
				call <- substitute(fun)
				fun <- match.fun(call[[1L]])
				params <- list( as.list(call[-1L]) )
			} else if ( identical(length(params), 0L) || !identical(mode(params[[1]]),"list") )
				params <- list(params)
			
			# function which saves variables at each program point
			.tracetag <- function (point=NULL) {
				if (!is.null(point)) {
					if (point<1) signalCondition(tracerCondition("The argument of .tracetag should be a natural number"))
					frame <- as.list(parent.frame())
					if (! is.null(vars)) { frame <- frame[vars]; names(frame)=vars }
					frame <- switch(noscalar,
							yes = frame,
							no =  Filter(function(x) { is.numeric(x) && identical(length(x),1L) }, frame),
							na = lapply(frame, function(x) {  if (is.numeric(x) && identical(length(x),1L)) x else NA_integer_ })
					)
					collectlist[[length(collectlist)+1L]] <<- c(list(.point=point),frame)	
				} 				
				currsteps <<- currsteps - 1L
				if (identical(currsteps,0L)) invokeRestart("timelimit")  
			}
			
			# create the sandbox environment where the function which will be executed			
			e <- new.env( parent=.GlobalEnv )
			assign(".tracetag",.tracetag, envir=e)		
			# the assume function is used in several examples 
			assign("assume", function(x) {}, envir=e)
			# the brandom function is used in serveral examples
			assign("brandom", function() { identical(rbinom(1, 1, 0.5), 1) }, envir=e)
			
			# annotate function to trace
			newf <- switch(tag,
					yes = annotate(fun, function() { currpoint <<- currpoint+1; bquote(.tracetag(.(currpoint))) }, skip=c(".tracetag"), ...),
					onlywhile = annotate(fun, function() { currpoint <<- currpoint+1; bquote(.tracetag(.(currpoint))) },  only.while = TRUE, skip=c(".tracetag"), ...),
					no = annotate(fun, quote(.tracetag()), only.while=TRUE, skip=c(".tracetag"), ...))			
			environment(newf) <- e
			
			# execute function and returns results
			for (i in 1:length(params)) {
				currsteps <- maxsteps
				withRestarts(do.call(newf,params[[i]]), "timelimit" = function() {})			
			}
			ProgramTrace$new(call = body(newf), trace = collectlist, originalvars = vars, noscalar = noscalar  )
		}
)

ProgramTrace <- proto(
		# call and trace are the program call and trace respectively
		# originalvars is the set of variables which has been traced (if specified manually)
		# noscalar is yes, no, na, according to programTrace
		# if originalvars is not null and noscalar is different from no, then we are sure that 
		# all the variables are similar at each program point
		new = function(., call, trace, originalvars, noscalar) {
			.$proto( call=call, trace=trace, originalvars=originalvars, noscalar=noscalar )
		},		
		getVars = function(.) {
			# memoize set of variables in ..vars
			if (! is.null(.[["vars"]])) return(.$..vars)
			if (! is.null(.$originalvars))
				.$..vars <- .$originalvars
			else if (! is.null(.[["..data"]]))
				.$..vars <- colnames(.$..data)			
			else
				.$..vars <- Reduce(union, Map(names, .$trace))
			.$..vars
		},
		getData = function(.) {
			# memoize tabular data in .$data
			if (! is.null(.[[".data"]])) return(.$..data)		
			# if noscalar is yes, it possible there are noscalar variables which should be turned into scalar
			if (identical(.$noscalar,"yes"))
				tmptrace <- lapply( .$trace, function(frame) lapply(frame, function(x) {  if (is.numeric(x) && identical(length(x),1L)) x else NA_integer_ }) )
			else
				tmptrace <- .$trace			
			if (identical(length(tmptrace),0L)) 
				.$..data <- as.data.frame(matrix(nrow=0L, ncol=1L+length(.$getVars()), dimnames=list( c(), c(".point",.$getVars())) ))				
			else  if (! is.null(.$originalvars) && !(identical(.$noscalar,"no")))
				.$..data <- as.data.frame( matrix(unlist(tmptrace), ncol=length(.$getVars())+1,
							byrow=TRUE,dimnames=list(NULL,c(".point",.$getVars()))) )
			else
				.$..data <- rbind.fill(tmptrace)
			.$..data
		},
		getCleanData = function (.) {
			# Remove columns which are entirely NA
			#data <- Filter(Negate(function(x) all(is.na(x),na.rm=FALSE)), .$getData())			

			# I prefer to convert NA to 0 insted of removing lines with NA, since this
			# may cause problems if a variable is introduced very late in the program
			# which is analyzed	
			#data[is.na(data)] <- 0L
			
			# Remove rows which contain some NA					
			data <- .$getData()
			if (nrow(data)==0)
				data
			else
				na.omit(data)
		},
		getCleanAggregatedData = function(.) {			
			agg <- list()
			data <- .$getCleanData()
			if (! empty(data)) {
				levels <- sort(unique(data$.point))
				for (x in levels)
					agg[[x]] <- data[data$.point == x, -1L]	
			}			
			agg
		}
)
